import timer


class AI(object):
	def __init__(self, type, scene):
		if type is 'Liquids':
			pass
		else:	#type == 'Solids'
			pass
			
		self.type = type
		self.timer = timer.Timer(150)
		self.pressed = False
		self.scene = scene
		
	def Update(self):
		pass

	def Render(self):
		pass
	
	"""returns the length of the shortest path, not the actual path."""
	def AStar(self, start, goal):
		closedSet = []								# The set of nodes already evaluated.	
		openSet = [start.tile]						 	# The set of tentative nodes to be evaluated.
		
		#using dictionaries for ease.
		g_score = {}
		h_score = {}
		f_score = {}
		g_score[start.tile] = 0							# Distance from start along optimal path.
		h_score[start.tile] = self.Heuristic(start.tile, goal)
		f_score[start.tile] = h_score[start.tile]		# Estimated total distance from start to goal through nodeB.
		came_from = {}
		self.passedGates = []
		while len(openSet) > 0:
			nodeA = openSet[0]
			for node in openSet:
				if f_score[node] < f_score[nodeA]:
					nodeA = node
			#nodeA = the node in openSet having the lowest f_score[] value
			
			if nodeA.x == goal.x and nodeA.y == goal.y:
				return self.ReconstructPathLength(came_from, nodeA)
			
			openSet.remove(nodeA)
			if FindTile(nodeA.x, nodeA.y, closedSet) is None:
				MyAddAndSort(closedSet, nodeA)
			
			for nodeB in nodeA.myTiles:
				if nodeB.wallFlag:
					continue
				if FindTile(nodeB.x, nodeB.y, closedSet):
					continue
					
				tentative_g_score = g_score[nodeA] + self.Heuristic(nodeA, nodeB)
 
				if FindTile(nodeB.x, nodeB.y, openSet) is None:
					MyAddAndSort(openSet, nodeB)
					tentative_is_better = True
				elif tentative_g_score < g_score[nodeB]:
					tentative_is_better = True
				else:
					tentative_is_better = False
					
				if tentative_is_better == True:
					came_from[nodeB] = nodeA
					g_score[nodeB] = tentative_g_score
					h_score[nodeB] = self.Heuristic(nodeB, goal)
					f_score[nodeB] = g_score[nodeB] + h_score[nodeB]
					
		return False, self.passedGates
 
	def ReconstructPathLength(self, came_from, current_node):
		if current_node in came_from:	#if this exists
			if current_node.gateFlag:
				self.passedGates.append(current_node)
			p, gates = self.ReconstructPathLength(came_from,came_from[current_node])
			return p + 1, self.passedGates
		else:
			return 0, self.passedGates
	
	def Heuristic(self, obj, goal):	#obj1 = start; obj2 = goal
		#h(n) = D * (abs(m.x-goal.x) + abs(n.y-goal.y))	# D = minimum cost from one tile to another
		return 1 * (abs(obj.x-goal.x) + abs(obj.y-goal.y))
	
class LiquidAI(AI):
	def __init__(self, scene):
		AI.__init__(self, "Liquids", scene)
		self.tapsOn = [None, None]
		self.count = 0
		self.maxCount = len(scene.taps)
		
		self.min1Obj = None
		self.min2Obj = None
		
	def Update(self):
		if not self.timer.IsDone():
			return
		
		self.timer.Reset()
		
		##horribly inefficient... whatever, unless it matters.
		#find two shortest paths with current settings
		
		min1 = 300
		min2 = 300
		
		if self.tapsOn[0]:
			for finish in self.scene.finishes:
				len, trash = self.AStar(self.tapsOn[0], finish)
				if len >= min1:
					continue
				min1 = len
		if self.tapsOn[1]:
			for finish in self.scene.finishes:
				len, trash = self.AStar(self.tapsOn[1], finish)
				if len >= min2:
					continue
				min2 = len
		
		while self.scene.taps[self.count] in self.tapsOn:
			self.count +=1
			if self.count >= self.maxCount:
				self.count = 0
				
		tap = self.scene.taps[self.count]
		for finish in self.scene.finishes:
			len, trash = self.AStar(tap, finish)
			if len >= min1 and len >= min2:
				continue
			
			#replace max between min1 and min2
			num = max(min1, min2)
			if num == min1 and self.min2Obj is not tap:
				min1 = len
				self.min1Obj = tap
			elif self.min1Obj is not tap:
				min2 = len
				self.min2Obj = tap
		
		##again, probably inefficient, but should work.
		#turn both off
		for tap in self.tapsOn:
			if tap:
				tap.ReceivedClick()
		#clear list and add new
		##allows None objects
		self.tapsOn = [self.min1Obj, self.min2Obj]
		#turn taps (back) on.
		for tap in self.tapsOn:
			if tap:
				tap.ReceivedClick()
		
		
		
		"""
		AI must monitor shortest distance from each tap to its nearest, unfilled, end-point.
		It should keep the shortest two paths "on."
		When a shorter path is revealed, either by the opening or closing of gates, that path should be used instead.
		
		This does not account for the fact that the water will have to fill up large rooms.	
		Will not be smart enough to run less than 2 taps.
		
		Should update ~1/second.
		"""
		
		
class SolidAI(AI):
	def __init__(self, scene):
		AI.__init__(self, 'Solids', scene)
		
		self.gatesOpen = []
		
	def Update(self):
		tapsOn = []
		for tap in self.scene.taps:
			if tap.on:
				tapsOn.append(tap)
			
		for gate in self.gatesOpen:
			gate.ToggleGate()
		
		self.gatesOpen = []
		
		
		min1 = 301
		min2 = 300
		gatesPassed = []
		for tap in tapsOn:
			if len(self.gatesOpen) >=2:
				break
			for finish in self.scene.finishes:
				if len(self.gatesOpen) >=2:
					break
				trash, gatesPassed = self.AStar(tap, finish)
				
				num = len(gatesPassed)
				if num is 0:
					continue
				else:
					self.gatesOpen.append(gatesPassed[num-1])
					
		if len(self.gatesOpen) < 2:
			for gate in gatesPassed:
				if not gate in self.gatesOpen:
					self.gatesOpen.append(gate)
					break
		for gate in self.gatesOpen:
			gate.ToggleGate()
			
		"""
		AI must monitor which taps are in use, and should attempt to enlongate the path their water must take.
		"""

"""
3 14
3 15
4 15
5 1
5 4
"""
def FindTile(x, y, tiles):
	i = 0
	xFound = False
	while i < len(tiles):
		if x is tiles[i].x:
			xFound = True
			break
		if x < tiles[i].x:
			break
		i += 1
	
	if not xFound:
		return None
	
	while i < len(tiles):
		if x != tiles[i].x:	
	
			break
		if y is tiles[i].y:
			assert x is tiles[i].x, "code thinks same tile--not true"
			return tiles[i]
		if y < tiles[i].y:
			break
		i += 1
		
	return None		
	
def MyAddAndSort(array, obj):
	if len(array) is 0:
		array.append(obj)
		return
	
	if array[0].x > obj.x:
		array.insert(0,obj)
		
	i = 0
	#from the front
	while i < len(array):
		if array[i].x == obj.x:
			break
		if array[i].x > obj.x:
			#i -= 1
			break
		i += 1
		
	while i < len(array):
		if array[i].x > obj.x:
			break
		if array[i].y > obj.y:
			break
		i += 1
	
	array.insert(i, obj)
	
	